#define M61_DISABLE 1
#include "m61.hh"
#include <cstdlib>
#include <cstring>
#include <cstdio>
#include <cinttypes>
#include <cassert>
#include <vector>
#include <algorithm>
#include <set>

struct metadata {  //added buffers so that struct size is 16 bytes
    size_t allocated_size; 
    bool is_free;
    char marker;
    const char* file; // work for test035, but where do I add (long) line?
    long line;
    char buffer[13];
    //struct metadata* next;
    //struct metadata* prev;
};
//metadata* head;
std::set<metadata*> metadata_set;

struct global_stats { //for tests 1-10
    unsigned long long numActive = 0;
    unsigned long long numTotal = 0;
    unsigned long long numFail = 0;
    unsigned long long activeByteSize = 0;
    unsigned long long totalByteSize = 0;
    unsigned long long failByteSize = 0;
    uintptr_t min_heap = -1;
    uintptr_t max_heap;
};

struct metachar {
    char boundary_marker;
};



static global_stats gstats; // for tests 1-10

/// m61_malloc(sz, file, line)
///    Return a pointer to `sz` bytes of newly-allocated dynamic memory.
///    The memory is not initialized. If `sz == 0`, then m61_malloc must
///    return a unique, newly-allocated pointer value. The allocation
///    request was at location `file`:`line`.

void* m61_malloc(size_t sz, const char* file, long line) {
    (void) file, (void) line;
    // Your code here.
    void* starting_pointer = base_malloc(sz + sizeof(metadata)+ sizeof(metachar));
    metadata* my_metadata = (metadata*) starting_pointer;
    void* payload_ptr = my_metadata + 1;
    //void* payload_ptr = (uintptr_t)begin_pointer + sizeof(size_t);
    //metadata* begin_pointer = base_malloc(sz + sizeof(metadata));
    
    if (sizeof(metadata) + sz < sz){ // test 5
        gstats.numFail++; // test 5
        gstats.failByteSize+=sz;
        return NULL; // test 5
        } // test 5

    if (starting_pointer == NULL){ // test 5
        gstats.numFail++; // test 5
        gstats.failByteSize+=sz;
        return NULL; // test 5
        }
    
    //what I originally had was: if (sz > 0 && ptr != NULL){ // test 2 pass & test 6/7 --> was not needed and excessive code. 

    // test 8 & 9 both max and min pointers. Use parentheses with type before variable name to cast to long int. 
    if ((uintptr_t) payload_ptr < gstats.min_heap || gstats.min_heap == -1 ){ // should I insert payload_ptr here for min and max?
        gstats.min_heap = (uintptr_t) payload_ptr;
    }
     // tests 8 & 9 both max and min pointers
    if (sz + (uintptr_t) payload_ptr > gstats.max_heap ){
        gstats.max_heap = (uintptr_t) payload_ptr + sz;
    }
    

    my_metadata->file = (const char*) file; // work for test035
    my_metadata->line = (long)line; // work for test035
    metachar* boundary_ptr = (metachar*)((char*)payload_ptr + sz);
    boundary_ptr->boundary_marker = (char)99;
    my_metadata->marker = 'Y';
    my_metadata->allocated_size = sz; 
    my_metadata->is_free=false;
    gstats.activeByteSize += my_metadata->allocated_size;
    gstats.numTotal++; // test 2 pass & tests 6/7
    gstats.numActive++;
    gstats.totalByteSize+=sz;
    metadata_set.insert (my_metadata);
    /* metadata* n = (uintptr_t*) my_metadata->allocated_size;
    my_metadata->next = head;
    my_metadata->prev = nullptr;
    if (head){
        head->prev = n;}
    head = n; */
    //printf("size is  %ld", sizeof(metadata));




    /*unsigned long long i = (uintptr_t)payload_ptr; //line 89-109 is for the linked list
    metadata* start = NULL;
    metadata* starter_pointer = (metadata)starting_pointer;
    //for (i= (uintptr_t)payload_ptr; i <= gstats.max_heap; i++)
    //starting_pointer = new metadata->memLeakCheck;
    starter_pointer->memLeakCheck= i;
    starter_pointer->next = NULL;
        if(start==NULL){
        start=starter_pointer;
        }
        if (start != NULL)
        {
        start= (metadata)payload_ptr;
        }*/
    //printf("size of the node is:  %ld", sizeof(Node));
    return payload_ptr; 
}

/// m61_free(ptr, file, line)
///    Free the memory space pointed to by `ptr`, which must have been
///    returned by a previous call to m61_malloc. If `ptr == NULL`,
///    does nothing. The free was called at location `file`:`line`.

void m61_free(void* ptr, const char* file, long line) {
    (void) file, (void) line;   // avoid uninitialized variable warnings
    // Your code here.
    if (ptr == NULL){ // test 7
        return; // test 7
    } 
    if ((uintptr_t)ptr <gstats.min_heap || (uintptr_t)ptr > gstats.max_heap){
        fprintf(stderr, "MEMORY BUG%s: invalid free of pointer %p, not in heap\n", file, ptr ); //different commands like %s, %ld, %p. Pass tests 21-24
        abort();
    }
    

    uintptr_t is_free_ptr = (uintptr_t)ptr - sizeof(metadata);
    metadata* free_metadata = (metadata*)is_free_ptr;
    metachar* boundary_ptr = (metachar*)(ptr + free_metadata->allocated_size) ;

    for (auto element : metadata_set){
        if (free_metadata->marker != 'Y'){
    fprintf(stderr, "MEMORY BUG: %s:%ld: invalid free of pointer %p, not allocated\n", file, line, ptr ); 
    fprintf(stderr,"%s:%ld: %p is %zu bytes inside a %zu byte region allocated here\n", element->file, element->line, ptr, element->marker + 39, element->allocated_size );
    abort();  
        }  
    }

    if (free_metadata->is_free){
    fprintf(stderr, "MEMORY BUG%s: invalid free of pointer %p, double free\n", file, ptr ); 
    abort();
    }
    if (boundary_ptr->boundary_marker != 99){
        fprintf(stderr, "MEMORY BUG%s: detected wild write during free of pointer %p\n", file, ptr ); 
        abort();
    }
    metadata_set.erase (free_metadata);
    free_metadata->is_free=true;
    gstats.numActive--;
    //metadata* temp = (metadata*) ptr;
    //is_free_ptr--;
    gstats.activeByteSize -= free_metadata->allocated_size;
    
  /* metadata* n = (metadata*) ptr;
    if (n->next){
        n->next->prev = n->prev;
    }
    if (n->prev){
        n->prev->next = n->next;
    }
    else
    {
        head = n->next;
    } */
    //ptr = NULL;
    //gstats.activeByteSize-=sz; // test 3 pass 
    //base_free(free_metadata); // changing this from ptr to temp makes me pass test 3&4
    //fprintf(stderr, "output is %d");

}

/// m61_calloc(nmemb, sz, file, line)
///    Return a pointer to newly-allocated dynamic memory big enough to
///    hold an array of `nmemb` elements of `sz` bytes each. If `sz == 0`,
///    then must return a unique, newly-allocated pointer value. Returned
///    memory should be initialized to zero. The allocation request was at
///    location `file`:`line`.

void* m61_calloc(size_t nmemb, size_t sz, const char* file, long line) {
    // Your code here (to fix test019).
    if (sz > ((size_t) -1)/nmemb){
        gstats.numFail++;
        gstats.failByteSize += sz*nmemb;
        return nullptr;
    }
    void* ptr = m61_malloc(nmemb * sz, file, line);
    
    if (ptr == nullptr) {
        memset(ptr, 0, nmemb * sz);
        gstats.numFail++;
        gstats.failByteSize += sz*nmemb;
        return nullptr;
    }

    return ptr;
}


/// m61_get_statistics(stats)
///    Store the current memory statistics in `*stats`.

void m61_get_statistics(m61_statistics* stats) {
    // Stub: set all statistics to enormous numbers
    memset(stats, 255, sizeof(m61_statistics));
    // Your code here.
    stats->nactive = gstats.numActive; // the -> does two operations: derefence a pointer (means currently you have an address to a memory spot, with arrow, tell computer go to that memory and grab object)
    stats->ntotal = gstats.numTotal; // class vs struct -> every obj in a struct is defaulted to a public object. A class defaults to private
    stats->nfail = gstats.numFail;
    stats->active_size = gstats.activeByteSize;
    stats->total_size = gstats.totalByteSize;
    stats->fail_size = gstats.failByteSize;
    stats->heap_max = gstats.max_heap; // test8
    stats->heap_min = gstats.min_heap; // test8
}


/// m61_print_statistics()
///    Print the current memory statistics.

void m61_print_statistics() {
    m61_statistics stats;
    m61_get_statistics(&stats);

    printf("alloc count: active %10llu   total %10llu   fail %10llu\n",
           stats.nactive, stats.ntotal, stats.nfail);
    printf("alloc size:  active %10llu   total %10llu   fail %10llu\n",
           stats.active_size, stats.total_size, stats.fail_size);
}


/// m61_print_leak_report()
///    Print a report of all currently-active allocated blocks of dynamic
///    memory.

void m61_print_leak_report() {
//void m61_print_leak_report(size_t sz, const char* file, long line) {
    // Your code here.
    // you can use char* that represents the file name bc it is constant. That string will never change. Bc file names stored globally. 
    for (auto element : metadata_set){
        fprintf(stdout, "LEAK CHECK: %s:%ld: allocated object %p with size %zu\n", element->file, element->line, (metadata*)element + 1, element->allocated_size);
    }
}


/// m61_print_heavy_hitter_report()
///    Print a report of heavily-used allocation locations.

void m61_print_heavy_hitter_report() {
    // Your heavy-hitters code here
}